"""
This file can process individual movie files in place.
There name, or directory name, will be analyzed, matched against
the porndb, and used for renaming (in place), and updating a mp4
file's metadata (poster, artists, etc.)
"""
import argparse
from dataclasses import dataclass
import pathlib
import string
import sys
from pathlib import Path
from random import choices
from typing import List, Optional

from loguru import logger

from namer.command import Command
from namer.comparison_results import ComparisonResult, ComparisonResults, LookedUpFileInfo
from namer.configuration import NamerConfig
from namer.configuration_utils import default_config, from_config, verify_configuration
from namer.command import make_command, move_command_files, move_to_final_location, set_permissions, write_log_file
from namer.ffmpeg import FFProbeResults, ffprobe
from namer.filenameparts import FileNameParts
from namer.metadataapi import get_complete_metadatapi_net_fileinfo, get_image, get_trailer, match
from namer.moviexml import parse_movie_xml_file, write_nfo
from namer.name_formatter import PartialFormatter
from namer.mutagen import update_mp4_file

DESCRIPTION = """
    Namer, the porndb local file renamer. It can be a command line tool to rename mp4/mkv/avi/mov/flv files and to embed tags in mp4s,
    or a watchdog service to do the above watching a directory for new files and moving matched files to a target location.
    File names are assumed to be of the form SITE.[YY]YY.MM.DD.String.of.performers.and.or.scene.name.<IGNORED_INFO>.[mp4|mkv].
    In the name, read the periods, ".", as any number of spaces " ", dashes "-", or periods ".".

    Provided you have an access token to the porndb (free sign up) https://www.metadataapi.net/, this program will
    attempt to match your file's name to search results from the porndb.   Please note that the site must at least be
    a substring of the actual site name on the porndb, and the date must be within one day or the release date on the
    porndb for a match to be considered.  If the log file flag is enabled then a <original file name minus ext>_namer.log
    file will be written with all the potential matches sorted, descending by how closely the scene name/performer names
    match the file's name segment after the 'SITE.[YY]YY.MM.DD'.
  """


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class ProcessingResults:
    """
    Returned from the namer.py process() function.   It contains information about if a match
    was found, and of so, where files were placed.  It also tracks if a directory was inputted
    to namer (rather than the exact movie file.)  That knowledge can be used to move directories
    and preserve relative files, or to delete left over artifacts.
    """

    search_results: Optional[List[ComparisonResult]] = None
    """
    True if a match was found in the porndb.
    """

    new_metadata: Optional[LookedUpFileInfo] = None
    """
    New metadata found for the file being processed.
    Sourced including queries against the porndb, which would be stored in search_results,
    or reading a .nfo xml file next to the video, with the file name identical except for
    the extension, which would be .nfo instead of .mp4,.mkv,.avi,.mov,.flv.
    """

    dir_file: Optional[Path] = None
    """
    Set if the input file for naming was a directory.   This has advantages, as clean up of other files is now possible,
    or all files can be moved to a destination specified in the field final_name_relative.
    """

    video_file: Optional[Path] = None
    """
    The location of the found video file.
    """

    parsed_file: Optional[FileNameParts] = None
    """
    The parsed file name.
    """

    final_name_relative: Optional[Path] = None
    """
    This is the full NamerConfig.new_relative_path_name string with all substitutions made.
    """


def dir_with_sub_dirs_to_process(dir_to_scan: Path, config: NamerConfig, infos: bool = False):
    """
    Used to find sub-dirs of a directory to be individually processed.
    The directories will be scanned for media and named/tagged in place
    based on config settings.
    """
    if dir_to_scan is not None and dir_to_scan.is_dir() and dir_to_scan.exists():
        logger.info("Scanning dir {} for sub-dirs/files to process", dir_to_scan)
        files = list(dir_to_scan.iterdir())
        files.sort()
        for file in files:
            fullpath_file = dir_to_scan / file
            if fullpath_file.is_dir() or fullpath_file.suffix.upper() in [".MP4", ".MKV"]:
                command = make_command(fullpath_file, config, nfo=infos, inplace=True)
                if command is not None:
                    process_file(command)


def tag_in_place(video: Optional[Path], config: NamerConfig, new_metadata: LookedUpFileInfo, ffprobe_results: Optional[FFProbeResults]):
    """
    Uses ComparisonResults to update a mp4 file's metadata based on a match in
    ComparisonResults.   Expects the first item of list to be the match if there is one.
    Will download a poster as well depending on NamerConfig config setting.
    """
    if new_metadata and video:
        poster = None
        if config.enabled_tagging and video.suffix.lower() == ".mp4":
            random = "".join(choices(population=string.ascii_uppercase + string.digits, k=10))
            poster = get_image(new_metadata.poster_url, random, video, config)
            logger.info("Updating file metadata (atoms): {}", video)
            update_mp4_file(video, new_metadata, poster, ffprobe_results, config)

        logger.info("Done tagging file: {}", video)
        if poster is not None and new_metadata.poster_url is not None and new_metadata.poster_url.startswith("http"):
            poster.unlink()


def get_local_metadata_if_requested(video_file: Path) -> Optional[LookedUpFileInfo]:
    """
    If there is an .nfo file next to the video_file, attempt to read it as
    an Emby/Jellyfin style movie xml file.
    """
    nfo_file = video_file.parent / (video_file.stem + ".nfo")
    if nfo_file.is_file() and nfo_file.exists():
        return parse_movie_xml_file(nfo_file)
    return None


def process_file(command: Command) -> Optional[Command]:
    """
    Bread and butter method.
    Given a file, determines if it's a dir, if so, the dir name may be used
    for comparison with the porndb.   The larges mp4/mkv file in the directory
    or any subdirectories will be assumed to be the movie file.

    Does not properly handle multipart movies.

    If the input file is not a dir it's name will be used, and it is assumed to
    be the movie file we wish to tag.

    The movie is either renamed in place if a file, or renamed and move to the root
    of the dir if a dir was passed in.

    The file is then update based on the metadata from the porndb if a mp4.
    """
    logger.info("Processing: {}", command.input_file)
    if command.target_movie_file is not None:
        new_metadata: Optional[LookedUpFileInfo] = None
        search_results: ComparisonResults = ComparisonResults([])
        # Match to nfo files, if enabled and found.
        if command.write_from_nfos:
            new_metadata = get_local_metadata_if_requested(command.target_movie_file)
            if new_metadata is not None:
                new_metadata.original_parsed_filename = command.parsed_file
            else:
                logger.error("""
                        Could not process files: {}
                        In the file's name should start with a site, a date and end with an extension""", command.input_file)
        # elif new_metadata is None and command.stashdb_id is not None and command.ff_probe_results is not None:
        #    phash = VideoPerceptualHash().get_phash(command.target_movie_file)
        #    todo use phash
        elif new_metadata is None and command.tpdb_id is not None and command.parsed_file is not None:
            file_infos = get_complete_metadatapi_net_fileinfo(command.parsed_file, command.tpdb_id, command.config)
            if file_infos is not None:
                new_metadata = file_infos
        elif new_metadata is None and command.parsed_file is not None and command.parsed_file.name is not None:
            search_results = match(command.parsed_file, command.config)
            if search_results:
                matched = search_results.get_match()
                if matched:
                    new_metadata = matched.looked_up
            if not command.target_movie_file:
                logger.error("""
                    Could not process file or directory: {}
                    Likely attempted to use the directory's name as the name to parse.
                    In general the dir or file's name should start with a site, a date and end with an extension
                    Target video file in dir was: {}""", command.input_file, command.target_movie_file)
        target_dir = command.target_directory if command.target_directory is not None else command.target_movie_file.parent
        set_permissions(target_dir, command.config)
        if new_metadata is not None:
            ffprobe_results = ffprobe(command.target_movie_file)
            target = move_to_final_location(command, new_metadata)
            tag_in_place(target.target_movie_file, command.config, new_metadata, ffprobe_results)
            add_extra_artifacts(target.target_movie_file, new_metadata, search_results, command.config)
            logger.info("Done processing file: {}, moved to {}", command.target_movie_file, target.target_movie_file)
            return target
        elif command.inplace is False:
            failed = move_command_files(command, command.config.failed_dir)
            if failed is not None and search_results is not None and failed.config.write_namer_failed_log:
                write_log_file(failed.target_movie_file, search_results, failed.config)
    return None


def add_extra_artifacts(video_file: Path, new_metadata: LookedUpFileInfo, search_results: ComparisonResults, config: NamerConfig):
    """
    Once the file is in its final location we will grab other relevant output if requested.
    """
    trailer = None
    if config.write_namer_log:
        write_log_file(video_file, search_results, config)

    if config.trailer_location and new_metadata is not None:
        trailer = get_trailer(new_metadata.trailer_url, video_file, config)

    if config.write_nfo and new_metadata is not None:
        poster = get_image(new_metadata.poster_url, "-poster", video_file, config)
        background = get_image(new_metadata.background_url, "-background", video_file, config)
        for performer in new_metadata.performers:
            poster = get_image(performer.image, performer.name.replace(" ", "-") + "-image", video_file, config)
            if poster is not None:
                performer.image = str(poster)

        write_nfo(video_file, new_metadata, config, trailer, poster, background)


def check_arguments(file_to_process: Path, dir_to_process: Path, config_override: Path):
    """
    check arguments.
    """
    error = False
    if file_to_process is not None:
        logger.info("File to process: {}", file_to_process)
        if not file_to_process.is_file() or not file_to_process.exists():
            logger.error("Error not a file! {}", file_to_process)
            error = True

    if dir_to_process is not None:
        logger.info("Directory to process: {}", dir_to_process)
        if not dir_to_process.is_dir() or not dir_to_process.exists():
            logger.info("Error not a directory! {}", dir_to_process)
            error = True

    if config_override is not None:
        logger.info("Config override specified: {}", config_override)
        if not config_override.is_file() or not config_override.exists():
            logger.info("Config override specified, but file does not exit: {}", config_override)
            error = True
    return error


def main(arg_list: List[str]):
    """
    Used to tag and rename files from the command line.
    See usage function above.
    """
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument("-c", "--configfile", type=pathlib.Path, help="config file, defaults first to env var NAMER_CONFIG, then local path namer.cfg, and finally ~/.namer.cfg.")
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-f", "--file", type=Path, help="a single file to process, and rename.")
    group.add_argument("-d", "--dir", type=Path, help="a directory to process.")
    parser.add_argument("-m", "--many", action="store_true", help="if set, a directory have all it's sub directories processed. Files move only within sub dirs, or are renamed in place, if in the root dir to scan")
    parser.add_argument("-i", "--infos", action="store_true", help="if set, .nfo files will attempt to be accessed next to movie files, if info files are found and parsed successfully, that metadata will be used rather than porndb matching. If using jellyfin .nfo files, please bump your release date by one day until they fix this issue: https://github.com/jellyfin/jellyfin/issues/7271.")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose, print logs")
    args = parser.parse_args(arg_list)
    level = "DEBUG" if args.verbose else "ERROR"
    logger.remove()
    logger.add(sys.stdout, format="{time} {level} {message}", level=level)
    check_arguments(args.file, args.dir, args.configfile)

    config = default_config()
    if args.configfile is not None and args.configfile.is_file():
        logger.info("Config override specified {}", args.configfile)
        config = from_config(args.configfile)
    verify_configuration(config, PartialFormatter())
    target = args.file
    if args.dir is not None:
        target = args.dir
    if args.many:
        dir_with_sub_dirs_to_process(args.dir.absolute(), config, args.infos)
    else:
        command = make_command(target.absolute(), config, inplace=True, nfo=args.infos)
        if command is not None:
            process_file(command)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
Helper functions to tie in to namer's functionality.
"""

import json
import math
import shutil
from pathlib import Path
from queue import Queue
from types import SimpleNamespace
from typing import Dict, List

from werkzeug.routing import Rule

from namer.configuration import NamerConfig
from namer.command import gather_target_files_from_dir, is_interesting_movie, subpath_or_equal, Command
from namer.metadataapi import __build_url, __get_response_json_object, __metadataapi_response_to_data  # type: ignore


def has_no_empty_params(rule: Rule) -> bool:
    """
    Currently unused, useful to inspect Flask rules.
    """
    defaults = rule.defaults if rule.defaults is not None else ()
    arguments = rule.arguments if rule.arguments is not None else ()
    return len(defaults) >= len(arguments)


def get_failed_files(config: NamerConfig) -> List[Dict]:
    """
    Get failed files to rename.
    """
    return list(map(command_to_file_info, gather_target_files_from_dir(config.failed_dir, config)))


def get_queued_files(queue: Queue, queue_limit: int = 100) -> List[Dict]:
    """
    Get queued files.
    """
    queue_items = list(queue.queue)[:queue_limit]
    return list(map(command_to_file_info, filter(lambda i: i is not None, queue_items)))


def get_queue_size(queue: Queue) -> int:
    return queue.qsize()


def command_to_file_info(command: Command) -> Dict:
    return {
        'file': str(command.target_movie_file.relative_to(command.config.failed_dir)) if subpath_or_equal(command.target_movie_file, command.config.failed_dir) else None,
        'name': command.target_directory.stem if command.parsed_dir_name and command.target_directory else command.target_movie_file.stem,
        'ext': command.target_movie_file.suffix[1:].upper(),
        'size_byte': command.target_movie_file.stat().st_size,
        'size': convert_size(command.target_movie_file.stat().st_size),
    }


def get_search_results(query: str, file: str, config: NamerConfig, performers_limit: int = 5) -> Dict:
    """
    Search results for user selection.
    """
    url = __build_url(config, name=query)
    json_response = __get_response_json_object(url, config)
    file_infos = []
    if json_response and json_response.strip() != '':
        json_obj = json.loads(json_response, object_hook=lambda d: SimpleNamespace(**d))
        formatted = json.dumps(json.loads(json_response), indent=4, sort_keys=True)
        file_infos = __metadataapi_response_to_data(json_obj, url, formatted, None)

    files = []
    for scene_data in file_infos:
        scene_performers = scene_data.performers
        other_performers = []
        if performers_limit:
            scene_performers = scene_data.performers[:performers_limit]
            other_performers = scene_data.performers[performers_limit:]

        performers = []
        for performer in scene_performers:
            performers.append({
                'name': performer.name,
                'gender': performer.role,
            })

        if other_performers:
            performers.append({
                'name': f'and {len(scene_data.performers) - performers_limit} more',
                'tooltip': ', '.join([performer.name for performer in other_performers])
            })

        scene = {
            'id': scene_data.uuid,
            'title': scene_data.name,
            'date': scene_data.date,
            'poster': scene_data.poster_url,
            'site': scene_data.site,
            'tags_count': len(scene_data.tags),
            'performers': performers,
        }
        files.append(scene)

    res = {
        'file': file,
        'files': files,
    }

    return res


def delete_file(file_name_str: str, config: NamerConfig) -> bool:
    """
    Delete selected file.
    """
    file_name = config.failed_dir / file_name_str
    if not is_acceptable_file(file_name, config) or not config.allow_delete_files:
        return False

    if config.del_other_files:
        target_name = config.failed_dir / Path(file_name_str).parts[0]
        shutil.rmtree(target_name)
    else:
        file_name.unlink()

    return not file_name.is_file()


def read_failed_log_file(name: str, config: NamerConfig) -> str:
    file_name = config.failed_dir / name
    file_name = file_name.parent / (file_name.stem + '_namer.log')
    data = ''
    if file_name.is_file():
        data = file_name.read_text('UTF-8')
    return data


def is_acceptable_file(file: Path, config: NamerConfig) -> bool:
    """
    Checks if a file belong to namer.
    """
    return str(config.failed_dir) in str(file.resolve()) and file.is_file() and is_interesting_movie(file, config)


def convert_size(size_bytes: int) -> str:
    """
    Convert int to size string.
    """
    if size_bytes == 0:
        return '0 B'

    size = 1024
    size_name = ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB')
    i = int(math.floor(math.log(size_bytes, size)))
    p = math.pow(size, i)
    s = round(size_bytes / p, 2)

    return f'{s:.2f} {size_name[i]}'


def human_format(num):
    if num == 0:
        return '0'

    size = 1000
    size_name = ('', 'K', 'M', 'B', 'T')
    i = int(math.floor(math.log(num, size)))
    p = math.pow(size, i)
    s = str(round(num / p, 2))
    s = s.rstrip('0').rstrip('.')

    return f'{s}{size_name[i]}'



"""
A file watching service to rename movie files and move them
to relevant locations after match the file against the porndb.
"""

import os
import re
import shutil
import sys
import tempfile
import time
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import List, Optional

import schedule
from loguru import logger
from watchdog.events import EVENT_TYPE_DELETED, EVENT_TYPE_MOVED, FileSystemEvent, PatternMatchingEventHandler
from watchdog.observers.polling import PollingObserver

from namer.configuration import NamerConfig
from namer.configuration_utils import default_config, verify_configuration
from namer.command import gather_target_files_from_dir, is_interesting_movie, make_command_relative_to, move_command_files, Command
from namer.name_formatter import PartialFormatter
from namer.namer import process_file
from namer.web.server import NamerWebServer


def done_copying(file: Optional[Path]) -> bool:
    """
    Determines if a file is being copied by checking its size in 2 second
    increments and seeing if the size has stayed the same.
    """
    if not file or not file.exists():
        return False

    while True:
        try:
            # pylint: disable=consider-using-with
            buffered_reader = open(file, mode="rb")
            buffered_reader.close()
            break
        except PermissionError:
            time.sleep(0.2)

    return True


@logger.catch
def handle(command: Command):
    """
    Responsible for processing and moving new movie files.
    """
    process_file(command)


def retry_failed(namer_config: NamerConfig):
    """
    Moves the contents from the failed dir to the watch dir to attempt reprocessing.
    """
    logger.info("Retry failed items:")

    # remove all old namer log files
    for log_file in namer_config.failed_dir.rglob("**/*_namer.log"):
        log_file.unlink()

    # move all files back to watch dir.
    for file in gather_target_files_from_dir(namer_config.failed_dir, namer_config):
        target = file.target_directory.name if file.parsed_dir_name and file.target_directory else file.target_movie_file.name
        shutil.move(namer_config.failed_dir / target, namer_config.watch_dir / target)


def is_fs_case_sensitive():
    """
    Create a temporary file to determine if a filesystem is case-sensitive, or not.
    """
    with tempfile.NamedTemporaryFile(prefix="TmP") as tmp_file:
        return not Path(tmp_file.name.lower()).is_file()


class MovieEventHandler(PatternMatchingEventHandler):
    """
    When a new movie file is detected, this class handles the event,
    and passes off the processing to the handler() function above.
    """

    __namer_config: NamerConfig

    def __init__(self, namer_config: NamerConfig, queue: Queue):
        super().__init__(patterns=["*.*"], case_sensitive=is_fs_case_sensitive(), ignore_directories=True, ignore_patterns=None)
        self.__namer_config = namer_config
        self.__command_queue = queue
        self.__re_ignored_dir = re.compile(namer_config.ignored_dir_regex)

    def on_any_event(self, event: FileSystemEvent):
        file_path = None
        if event.event_type == EVENT_TYPE_MOVED:
            file_path = event.dest_path  # type: ignore
        elif event.event_type != EVENT_TYPE_DELETED:
            file_path = event.src_path

        if file_path:
            path = Path(file_path)
            relative_path = str(path.relative_to(self.__namer_config.watch_dir))
            if not self.__re_ignored_dir.search(relative_path) and done_copying(path) and is_interesting_movie(path, self.__namer_config):
                logger.info("watchdog process called for {}", relative_path)

                # Extra wait time in case other files are copies in as well.
                if self.__namer_config.del_other_files:
                    time.sleep(self.__namer_config.extra_sleep_time)

                self.prepare_file_for_processing(path)

    @logger.catch
    def prepare_file_for_processing(self, path: Path):
        command = make_command_relative_to(input_dir=path, relative_to=self.__namer_config.watch_dir, config=self.__namer_config)
        working_command = move_command_files(command, self.__namer_config.work_dir)
        if working_command is not None:
            self.__command_queue.put(working_command)


class MovieWatcher:
    """
    Watches a configured dir for new files and attempts
    to process them and move to a destination dir.

    If a failure occurs moves the new files to a failed dir.

    See NamerConfig
    """

    def __processing_thread(self):
        while True:
            command = self.__command_queue.get()
            if command is None:
                break

            handle(command)
            self.__command_queue.task_done()
        self.__command_queue.task_done()

    def __init__(self, namer_config: NamerConfig):
        self.__started = False
        self.__stopped = False
        self.__namer_config = namer_config
        self.__src_path = namer_config.watch_dir
        self.__event_observer = PollingObserver()
        self.__webserver: Optional[NamerWebServer] = None
        self.__command_queue: Queue = Queue()
        self.__worker_thread: Thread = Thread(target=self.__processing_thread, daemon=True)
        self.__event_handler = MovieEventHandler(namer_config, self.__command_queue)
        self.__background_thread: Optional[Thread] = None

    def run(self):
        """
        Checks for new files in 3 second intervals,
        needed if running in docker as events aren't properly passed in.
        """
        if not self.__started:
            self.start()
            if self.__namer_config.web:
                self.__webserver = NamerWebServer(self.__namer_config, self.__command_queue)
                self.__webserver.start()

            try:
                while not self.__stopped:
                    schedule.run_pending()
                    time.sleep(3)
                self.stop()
            except KeyboardInterrupt:
                self.stop()

            self.__started = False
            self.__stopped = False

    def __enter__(self):
        self.__background_thread = Thread(target=self.run)
        self.__background_thread.start()

        tries = 0
        while not self.get_web_port() and tries < 20:
            time.sleep(0.2)
            tries += 1

        if not self.get_web_port:
            raise RuntimeError("application did not get assigned a port within 4 seconds.")

        return self

    def __simple_exit__(self):
        self.stop()
        if self.__background_thread:
            self.__background_thread.join()
            self.__background_thread = None

    def __exit__(self, exc_type, exc_value, traceback):
        self.__simple_exit__()

    def start(self):
        """
        starts a background thread to check for files.
        """
        config = self.__namer_config
        logger.info("Start porndb scene watcher.... watching: {}", config.watch_dir)

        if os.environ.get("PROJECT_VERSION"):
            project_version = os.environ.get("PROJECT_VERSION")
            print(f"Namer version: {project_version}")

        if os.environ.get("BUILD_DATE"):
            build_date = os.environ.get("BUILD_DATE")
            print(f"Built on: {build_date}")

        if os.environ.get("GIT_HASH"):
            git_hash = os.environ.get("GIT_HASH")
            print(f"Git Hash: {git_hash}")

        self.__schedule()
        self.__event_observer.start()
        self.__worker_thread.start()

        # touch all existing movie files.
        files: List[Path] = []
        for file in self.__namer_config.watch_dir.rglob("**/*.*"):
            if file.is_file() and file.suffix.lower()[1:] in self.__namer_config.target_extensions:
                files.append(file)

        for file in files:
            if file.exists() and file.is_file():
                self.__event_handler.prepare_file_for_processing(file)

    def stop(self):
        """
        stops a background thread to check for files.
        Waits for the event processor to complete.
        """
        if not self.__stopped:
            self.__stopped = True
            logger.info("exiting")
            self.__event_observer.stop()
            self.__event_observer.join()
            if self.__webserver:
                self.__webserver.stop()
            self.__command_queue.put(None)
            self.__command_queue.join()
            logger.info("exited")

    def get_web_port(self) -> Optional[int]:
        if self.__webserver is not None:
            return self.__webserver.get_effective_port()

    def __schedule(self):
        self.__event_observer.schedule(self.__event_handler, str(self.__src_path), recursive=True)


def create_watcher(namer_watchdog_config: NamerConfig) -> MovieWatcher:
    """
    Configure and start a watchdog looking for new Movies.
    """
    logger.remove()
    logger.add(sys.stdout, format="{time} {level} {message}", level="INFO", diagnose=False)
    logger.info(namer_watchdog_config)

    if not verify_configuration(namer_watchdog_config, PartialFormatter()):
        sys.exit(-1)

    if namer_watchdog_config.retry_time:
        schedule.every().day.at(namer_watchdog_config.retry_time).do(lambda: retry_failed(namer_watchdog_config))

    movie_watcher = MovieWatcher(namer_watchdog_config)

    return movie_watcher


if __name__ == "__main__":
    create_watcher(default_config()).run()



"""
Parse string in to FileNamePart define in namer_types.
"""
from dataclasses import dataclass
import re
from pathlib import PurePath
from typing import Optional, Pattern

from loguru import logger


DEFAULT_REGEX_TOKENS = "{_site}{_sep}{_optional_date}{_ts}{_name}{_dot}{_ext}"


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class FileNameParts:
    """
    Represents info parsed from a file name, usually of a nzb, named something like:
    'EvilAngel.22.01.03.Carmela.Clutch.Fabulous.Anal.3-Way.XXX.2160p.MP4-GAYME-xpost'
    or
    'DorcelClub.20.12..Aya.Benetti.Megane.Lopez.And.Bella.Tina.2160p.MP4-GAYME-xpost'
    """

    # pylint: disable=too-many-instance-attributes

    site: Optional[str] = None
    """
    Site the file originated from, "DorcelClub", "EvilAngel", etc.
    """
    date: Optional[str] = None
    """
    formatted: YYYY-mm-dd
    """
    trans: bool = False
    """
    If the name originally started with an "TS" or "ts"
    it will be stripped out and placed in a separate location, aids in matching, usable to genre mark content.
    """
    name: Optional[str] = None
    """
    The remained of a file, usually between the date and video markers such as XXX, 4k, etc.   Heavy lifting
    occurs to match this to a scene name, perform names, or a combo of both.
    """
    extension: Optional[str] = None
    """
    The file's extension .mp4 or .mkv
    """
    source_file_name: Optional[str] = None
    """
    What was originally parsed.
    """

    def __str__(self) -> str:
        return f"""site: {self.site}
        date: {self.date}
        trans: {self.trans}
        name: {self.name}
        extension: {self.extension}
        original full name: {self.source_file_name}
        """


def name_cleaner(name: str) -> str:
    """
    Given the name parts, following a date, but preceding the file extension, attempt to glean
    extra information and discard useless information for matching with the porndb.
    """
    # truncating cruft
    for size in ["2160p", "1080p", "720p", "4k", "3840p"]:
        name = re.sub(r"[.\- ]" + size + r"[.\- ]?.*", "", name)
    # remove trailing ".XXX."
    name = re.sub(r"[.\- ]?XXX[.\- ]?.*$", "", name)
    name = re.sub(r"\.", " ", name)
    return name


def parser_config_to_regex(tokens: str) -> Pattern[str]:
    """
    ``{_site}{_sep}{_optional_date}{_ts}{_name}{_dot}{_ext}``

    ``Site - YYYY.MM.DD - TS - name.mkv``

    ```
    _sep            r'[\\.\\- ]+'
    _site           r'(?P<site>[a-zA-Z0-9\\'\\.\\-\\ ]*?[a-zA-Z0-9]*?)'
    _date           r'(?P<year>[0-9]{2}(?:[0-9]{2})?)[\\.\\- ]+(?P<month>[0-9]{2})[\\.\\- ]+(?P<day>[0-9]{2})'
    _optional_date  r'(?:(?P<year>[0-9]{2}(?:[0-9]{2})?)[\\.\\- ]+(?P<month>[0-9]{2})[\\.\\- ]+(?P<day>[0-9]{2})[\\.\\- ]+)?'
    _ts             r'((?P<trans>[T|t][S|s])'+_sep+'){0,1}'
    _name           r'(?P<name>(?:.(?![0-9]{2,4}[\\.\\- ][0-9]{2}[\\.\\- ][0-9]{2}))*)'
    _dot            r'\\.'
    _ext            r'(?P<ext>[a-zA-Z0-9]{3,4})$'
    ```
    """

    _sep = r"[\.\- ]+"
    _site = r"(?P<site>.*?)"
    _date = r"(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})"
    _optional_date = r"(?:(?P<year>[0-9]{2}(?:[0-9]{2})?)[\.\- ]+(?P<month>[0-9]{2})[\.\- ]+(?P<day>[0-9]{2})[\.\- ]+)?"
    _ts = r"((?P<trans>[T|t][S|s])" + _sep + "){0,1}"
    _name = r"(?P<name>(?:.(?![0-9]{2,4}[\.\- ][0-9]{2}[\.\- ][0-9]{2}))*)"
    _dot = r"\."
    _ext = r"(?P<ext>[a-zA-Z0-9]{3,4})$"
    regex = tokens.format_map(
        {
            "_site": _site,
            "_date": _date,
            "_optional_date": _optional_date,
            "_ts": _ts,
            "_name": _name,
            "_ext": _ext,
            "_sep": _sep,
            "_dot": _dot,
        }
    )
    return re.compile(regex)


def parse_file_name(filename: str, regex_config: str = DEFAULT_REGEX_TOKENS) -> FileNameParts:
    """
    Given an input name of the form site-yy.mm.dd-some.name.part.1.XXX.2160p.mp4,
    parses out the relevant information in to a structure form.
    """
    regex = parser_config_to_regex(regex_config)
    file_name_parts = FileNameParts()
    file_name_parts.extension = PurePath(filename).suffix[1:]
    match = regex.search(filename)
    if match:
        if match.groupdict().get("year"):
            prefix = "20" if len(match.group("year")) == 2 else ""
            file_name_parts.date = prefix + match.group("year") + "-" + match.group("month") + "-" + match.group("day")

        if match.groupdict().get("name"):
            file_name_parts.name = name_cleaner(match.group("name"))

        if match.groupdict().get("site"):
            file_name_parts.site = match.group("site")

        if match.groupdict().get("trans"):
            trans = match.group("trans")
            file_name_parts.trans = bool(trans and trans.strip().upper() == "TS")

        file_name_parts.extension = match.group("ext")
        file_name_parts.source_file_name = filename
    else:
        logger.warning("Could not parse target name which may be a file (or directory) name depending on settings and input: {}", filename)

    return file_name_parts



"""
Defines the web routes of a Flask webserver for namer.
"""
from queue import Queue

from flask import Blueprint, redirect, render_template
from flask.wrappers import Response

from namer.configuration import NamerConfig
from namer.web.actions import get_failed_files, get_queued_files


def get_routes(config: NamerConfig, command_queue: Queue) -> Blueprint:
    """
    Builds a blueprint for flask with passed in context, the NamerConfig.
    """
    blueprint = Blueprint('web', __name__)

    """
    @blueprint.route('/')
    def index() -> str:
         data = []
         for rule in app.url_map.iter_rules():
             if rule.methods is not None and 'GET' in rule.methods and has_no_empty_params(rule):
                 url = url_for(rule.endpoint, **(rule.defaults or {}))
                 data.append((url, rule.endpoint))

         return render_template('pages/index.html', links=data)
    """

    @blueprint.route('/')
    def index() -> Response:
        return redirect('failed', code=302)  # type: ignore

    @blueprint.route('/failed')
    def failed() -> str:
        """
        Displays all failed to name files.
        """
        data = get_failed_files(config)
        return render_template('pages/failed.html', data=data, config=config)

    @blueprint.route('/queue')
    def queue() -> str:
        """
        Displays all queued files.
        """
        data = get_queued_files(command_queue)
        return render_template('pages/queue.html', data=data, config=config)

    @blueprint.route('/settings')
    def settings() -> str:
        """
        Displays namer settings.
        """
        return render_template('pages/settings.html', config=config)

    return blueprint



import concurrent.futures
import subprocess
import platform
import zipfile
from decimal import Decimal, ROUND_HALF_UP
from pathlib import Path
from typing import List, Literal, Optional

import imagehash
import numpy
import scipy.fft
import scipy.fftpack
from loguru import logger
from PIL import Image

from namer.ffmpeg import extract_screenshot, ffprobe
from namer.http import Http


class VideoPerceptualHash:
    __screenshot_width: int = 160
    __columns: int = 5
    __rows: int = 5

    __phash_path: Optional[Path]
    __phash_name: str = 'stash_phash'

    def __init__(self):
        self.__phash_path = Path(__file__).parent.parent / 'tools'
        if not self.__phash_path.is_dir():
            self.__phash_path.mkdir(exist_ok=True, parents=True)

        if not [file for file in self.__phash_path.glob('*') if self.__phash_name == file.stem]:
            self.__download_stash_phash()

    def get_phash(self, file: Path) -> Optional[imagehash.ImageHash]:
        phash = None

        thumbnail_image = self.__generate_image_thumbnail(file)
        if thumbnail_image:
            phash = self.__phash(thumbnail_image, hash_size=8, high_freq_factor=8, resample=Image.Resampling.BILINEAR)  # type: ignore

        return phash

    def __generate_image_thumbnail(self, file: Path) -> Optional[Image.Image]:
        thumbnail_image = None

        probe = ffprobe(file)
        if not probe:
            return thumbnail_image

        duration = probe.get_format().duration

        thumbnail_list = self.__generate_thumbnails(file, duration)
        if thumbnail_list:
            thumbnail_image = self.__concat_images(thumbnail_list)

        return thumbnail_image

    def get_stash_phash(self, file: Path) -> Optional[imagehash.ImageHash]:
        return self.__execute_stash_phash(file)

    def __download_stash_phash(self):
        os = platform.system().lower()
        url = f'https://github.com/DirtyRacer1337/stash_phash/releases/download/nightly/stash_phash-{os}.zip'
        http_file = Http.download_file(url)
        if http_file:
            zipfile.ZipFile(http_file).extractall(self.__phash_path)
            if os != 'windows' and self.__phash_path:
                file = self.__phash_path / self.__phash_name
                file.chmod(0o777)

        return bool(http_file)

    def __execute_stash_phash(self, file: Path) -> Optional[imagehash.ImageHash]:
        phash = None
        if not self.__phash_path:
            return phash

        args = [
            str(self.__phash_path / self.__phash_name),
            '-f', str(file),
        ]
        with subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True) as process:
            stdout, stderr = process.communicate()
            stdout, stderr = stdout.strip(), stderr.strip()
            success = process.returncode == 0
            if success:
                phash = imagehash.hex_to_hash(stdout)
            else:
                logger.error(stderr)

        return phash

    def __generate_thumbnails(self, file: Path, duration: float) -> List[Image.Image]:
        duration = int(Decimal(duration * 100).quantize(0, ROUND_HALF_UP)) / 100

        chunk_count = self.__columns * self.__rows
        offset = 0.05 * duration
        step_size = (0.9 * duration) / chunk_count

        if duration / chunk_count < 0.03:
            return []

        queue = []
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for idx in range(chunk_count):
                time = offset + (idx * step_size)
                future = executor.submit(extract_screenshot, file, time, self.__screenshot_width)
                queue.append(future)

        concurrent.futures.wait(queue)
        images = [item.result() for item in queue]

        return images

    def __concat_images(self, images: List[Image.Image]) -> Image.Image:
        width, height = images[0].size

        image_size = (width * self.__columns, height * self.__rows)
        image = Image.new('RGB', image_size)

        for row in range(self.__rows):
            for col in range(self.__columns):
                offset = width * col, height * row
                idx = row * self.__columns + col
                image.paste(images[idx], offset)

        return image

    @staticmethod
    def __phash(image: Image.Image, hash_size=8, high_freq_factor=4, resample: Literal[0, 1, 2, 3, 4, 5] = Image.Resampling.LANCZOS) -> Optional[imagehash.ImageHash]:  # type: ignore
        if hash_size < 2:
            raise ValueError("Hash size must be greater than or equal to 2")

        img_size = hash_size * high_freq_factor
        image = image.resize((img_size, img_size), resample).convert('L')
        pixels = numpy.asarray(image)

        dct = scipy.fft.dct(scipy.fft.dct(pixels, axis=0), axis=1)
        dct_low_freq = dct[:hash_size, :hash_size]  # type: ignore
        med = numpy.median(dct_low_freq)
        diff = dct_low_freq > med

        return imagehash.ImageHash(diff)



"""
A wrapper allowing shutdown of a Flask server.
"""
from asyncio.log import logger
import mimetypes
from queue import Queue
from threading import Thread
from typing import List, Optional, Union

from flask import Flask, Blueprint
from flask_compress import Compress
from waitress import create_server
from waitress.server import BaseWSGIServer, MultiSocketServer

from namer.configuration import NamerConfig
from namer.web.routes import api, web


class GenericWebServer:
    """
    A wrapper allowing shutdown of a Flask server.
    """
    __app: Flask
    __compress = Compress()

    __port: int
    __host: str
    __path: str
    __blueprints: List[Blueprint]

    __server: Union[MultiSocketServer, BaseWSGIServer]
    __thread: Thread

    __mime_types: dict = {
        'font/woff': '.woff',
        'font/woff2': '.woff2',
    }

    def __init__(self, host: str, port: int, webroot: Optional[str], blueprints: List[Blueprint], static_path: Optional[str] = 'public'):
        self.__host = host
        self.__port = port
        self.__path = '/' if not webroot else webroot
        self.__app = Flask(__name__, static_url_path=self.__path, static_folder=static_path, template_folder='templates')
        self.__blueprints = blueprints

        self.__add_mime_types()
        self.__register_blueprints()
        self.__make_server()

    def __make_server(self):
        self.__compress.init_app(self.__app)
        self.__server = create_server(self.__app, host=self.__host, port=self.__port, clear_untrusted_proxy_headers=True)
        self.__thread = Thread(target=self.__run, daemon=True)

    def __register_blueprints(self):
        for blueprint in self.__blueprints:
            blueprint_path = self.__path + blueprint.url_prefix if blueprint.url_prefix else self.__path
            self.__app.register_blueprint(blueprint, url_prefix=blueprint_path)

    def __add_mime_types(self):
        self.__app.json.mimetype = 'application/json; charset=utf-8'  # type: ignore

        for mime, ext in self.__mime_types.items():
            test_mime, test_ext = mimetypes.guess_type(f'0{ext}')
            if test_mime is None:
                mimetypes.add_type(mime, ext)

    def start(self):
        self.__thread.start()

    def __run(self):
        """
        Start server on existing thread.
        """
        if self.__server:
            try:
                self.__server.run()
            except OSError:
                logger.error("Stopping server")
            finally:
                self.stop()

    def stop(self):
        """
        Stop severing requests and empty threads.
        """
        if self.__server:
            self.__server.close()

    def get_effective_port(self) -> Optional[int]:
        return getattr(self.__server, "effective_port", None)

    def get_url(self) -> str:
        """
        Returns the full url to access this server, usually http://127.0.0.1:<os assigned port>/
        """
        return f"http://{self.__host}:{self.get_effective_port()}{self.__path}"


class NamerWebServer(GenericWebServer):
    __namer_config: NamerConfig
    __command_queue: Queue

    def __init__(self, namer_config: NamerConfig, command_queue: Queue):
        self.__namer_config = namer_config
        self.__command_queue = command_queue
        webroot = "/" if not self.__namer_config.web_root else self.__namer_config.web_root
        blueprints = [
            web.get_routes(self.__namer_config, self.__command_queue),
            api.get_routes(self.__namer_config, self.__command_queue),
        ]

        super().__init__(self.__namer_config.host, self.__namer_config.port, webroot, blueprints)



"""
Handle matching data in a FileNameParts (likely generated by a namer_file_parser.py) to
look up metadata (actors, studio, creation data, posters, etc) from the porndb.
"""

import argparse
import itertools
import json
import re
import sys
from datetime import date, timedelta
from pathlib import Path
from types import SimpleNamespace
from typing import List, Optional, Tuple
from urllib.parse import quote

import rapidfuzz
from loguru import logger
from PIL import Image
from unidecode import unidecode

from namer.comparison_results import ComparisonResult, ComparisonResults, LookedUpFileInfo, Performer
from namer.configuration import NamerConfig
from namer.configuration_utils import default_config
from namer.command import make_command, set_permissions
from namer.filenameparts import FileNameParts
from namer.http import Http


def __find_best_match(query: Optional[str], match_terms: List[str], config: NamerConfig) -> Tuple[str, float]:
    powerset_iter = []

    max_size = min(len(match_terms), config.max_performer_names)
    for length in range(1, max_size + 1):
        data = map(" ".join, itertools.combinations(match_terms, length))
        powerset_iter = itertools.chain(powerset_iter, data)

    ratio = rapidfuzz.process.extractOne(query, choices=powerset_iter)
    return (ratio[0], ratio[1]) if ratio else ratio


def __attempt_better_match(existing: Tuple[str, float], query: Optional[str], match_terms: List[str], namer_config: NamerConfig) -> Tuple[str, float]:
    if existing and existing[1] >= 89.9:  # magic numer
        return existing

    found = __find_best_match(query, match_terms, namer_config)
    if not existing:
        return found

    if not found:
        return "", 0.0

    return existing if existing[1] >= found[1] else found


def __evaluate_match(name_parts: FileNameParts, looked_up: LookedUpFileInfo, namer_config: NamerConfig) -> ComparisonResult:
    site = False
    found_site = None
    if looked_up.site:
        found_site = re.sub(r"[^a-z0-9]", "", looked_up.site.lower())
        if not name_parts.site:
            site = True
        else:
            site = re.sub(r"[^a-z0-9]", "", name_parts.site.lower()) in found_site or re.sub(r"[^a-z0-9]", "", unidecode(name_parts.site.lower())) in found_site

    if found_site in namer_config.sites_with_no_date_info:
        release_date = True
    else:
        release_date = bool(name_parts.date and (name_parts.date == looked_up.date or unidecode(name_parts.date) == looked_up.date))

    result: Tuple[str, float] = ('', 0.0)

    # Full Name
    all_performers = list(map(lambda p: p.name, looked_up.performers))
    if looked_up.name:
        all_performers.insert(0, looked_up.name)

    result = __attempt_better_match(result, name_parts.name, all_performers, namer_config)
    if name_parts.name:
        result = __attempt_better_match(result, unidecode(name_parts.name), all_performers, namer_config)

    # First Name Powerset.
    if result and result[1] < 89.9:
        all_performers = list(map(lambda p: p.name.split(" ")[0], looked_up.performers))
        if looked_up.name:
            all_performers.insert(0, looked_up.name)
        result = __attempt_better_match(result, name_parts.name, all_performers, namer_config)
        if name_parts.name:
            result = __attempt_better_match(result, unidecode(name_parts.name), all_performers, namer_config)

    return ComparisonResult(
        name=result[0],
        name_match=result[1],
        date_match=release_date,
        site_match=site,
        name_parts=name_parts,
        looked_up=looked_up,
    )


def __update_results(results: List[ComparisonResult], name_parts: FileNameParts, namer_config: NamerConfig, skip_date: bool = False, skip_name: bool = False):
    if not results or not results[0].is_match():
        for match_attempt in __get_metadataapi_net_fileinfo(name_parts, namer_config, skip_date, skip_name):
            result = __evaluate_match(name_parts, match_attempt, namer_config)
            results.append(result)
        results = sorted(results, key=__match_percent, reverse=True)

    return results


def __metadata_api_lookup(name_parts: FileNameParts, namer_config: NamerConfig) -> List[ComparisonResult]:
    results = []
    results = __update_results(results, name_parts, namer_config)
    results = __update_results(results, name_parts, namer_config, skip_date=True)
    results = __update_results(results, name_parts, namer_config, skip_date=True, skip_name=True)
    results = __update_results(results, name_parts, namer_config, skip_name=True)

    if name_parts.date and (not results or not results[-1].is_match()):
        name_parts.date = (date.fromisoformat(name_parts.date) + timedelta(days=-1)).isoformat()
        logger.info("Not found, trying 1 day before: {}", name_parts)
        results = __update_results(results, name_parts, namer_config)
        results = __update_results(results, name_parts, namer_config, skip_date=False, skip_name=True)

    if name_parts.date and (not results or not results[-1].is_match()):
        name_parts.date = (date.fromisoformat(name_parts.date) + timedelta(days=2)).isoformat()
        logger.info("Not found, trying 1 day after: {}", name_parts)
        results = __update_results(results, name_parts, namer_config)
        results = __update_results(results, name_parts, namer_config, skip_date=False, skip_name=True)

    return results


def __match_percent(result: ComparisonResult) -> float:
    add_value = 0.00
    if result.is_match():
        add_value = 1000.00

    value = (result.name_match + add_value) if result and result.name_match else add_value
    logger.debug("Name match was {:.2f} for {}", value, result.name)

    return value


@logger.catch
def __get_response_json_object(url: str, config: NamerConfig) -> str:
    """
    returns json object with info
    """
    headers = {
        "Authorization": f"Bearer {config.porndb_token}",
        "Content-Type": "application/json",
        "Accept": "application/json",
        "User-Agent": "namer-1",
    }
    with Http.get(url, cache_session=config.cache_session, headers=headers) as response:
        response.raise_for_status()
        return response.text


@logger.catch
def download_file(url: str, file: Path, config: NamerConfig) -> bool:
    headers = {
        "User-Agent": "namer-1",
    }
    if "metadataapi.net" in url:
        headers["Authorization"] = f"Bearer {config.porndb_token}"

    http_file = Http.download_file(url, headers=headers)
    if http_file:
        file.write_bytes(http_file.getbuffer().tobytes())

    return bool(http_file)


@logger.catch
def get_image(url: Optional[str], infix: str, video_file: Optional[Path], config: NamerConfig) -> Optional[Path]:
    """
    returns json object with info
    """
    if url and video_file:
        file = video_file.parent / (video_file.stem + infix + '.png')
        if config.enabled_poster and url.startswith("http") and not file.exists():
            file.parent.mkdir(parents=True, exist_ok=True)
            if download_file(url, file, config):
                with Image.open(file) as img:
                    img.save(file, 'png')
                set_permissions(file, config)
                return file
            else:
                return None

        poster = (video_file.parent / url).resolve()
        return poster if poster.exists() and poster.is_file() else None


@logger.catch
def get_trailer(url: Optional[str], video_file: Optional[Path], namer_config: NamerConfig) -> Optional[Path]:
    """
    returns json object with info
    """
    if namer_config.trailer_location and url and video_file:
        logger.info("Attempting to download trailer")
        location = namer_config.trailer_location[:max([idx for idx, x in enumerate(namer_config.trailer_location) if x == "."])]
        url_parts = url.split("?")[0].split(".")

        ext = "mp4"
        if url_parts and url_parts[-1].lower() in namer_config.target_extensions:
            ext = url_parts[-1]

        trailer_file: Path = video_file.parent / (location + "." + ext)
        trailer_file.parent.mkdir(parents=True, exist_ok=True)
        if not trailer_file.exists() and url.startswith("http"):
            if download_file(url, trailer_file, namer_config):
                set_permissions(trailer_file, namer_config)
                return trailer_file
            else:
                return None

        trailer = (video_file.parent / url).resolve()
        return trailer if trailer.exists() and trailer.is_file() else None


def __json_to_fileinfo(data, url, json_response, name_parts) -> LookedUpFileInfo:
    file_info = LookedUpFileInfo()
    file_info.uuid = data._id  # pylint: disable=protected-access
    file_info.name = data.title
    file_info.description = data.description
    file_info.date = data.date
    file_info.source_url = data.url
    file_info.poster_url = data.poster
    file_info.trailer_url = data.trailer
    if data.background:
        file_info.background_url = data.background.large
    file_info.site = data.site.name
    file_info.look_up_site_id = data._id  # pylint: disable=protected-access

    for json_performer in data.performers:
        if not json_performer.name:
            continue
        performer = Performer(json_performer.name)
        if hasattr(json_performer, "parent") and hasattr(json_performer.parent, "extras"):
            performer.role = json_performer.parent.extras.gender
        performer.image = json_performer.image
        file_info.performers.append(performer)

    file_info.original_query = url
    file_info.original_response = json_response
    file_info.original_parsed_filename = name_parts

    tags = []
    if hasattr(data, "tags"):
        for tag in data.tags:
            tags.append(tag.name)
        file_info.tags = tags

    return file_info


def __metadataapi_response_to_data(json_object, url, json_response, name_parts) -> List[LookedUpFileInfo]:
    file_infos = []
    if hasattr(json_object, "data"):
        if isinstance(json_object.data, list):
            for data in json_object.data:
                found_file_info = __json_to_fileinfo(data, url, json_response, name_parts)
                file_infos.append(found_file_info)
        else:
            found_file_info = __json_to_fileinfo(json_object.data, url, json_response, name_parts)
            file_infos.append(found_file_info)

    return file_infos


def __build_url(namer_config: NamerConfig, site: Optional[str] = None, release_date: Optional[str] = None, name: Optional[str] = None, uuid: Optional[str] = None) -> str:
    if uuid:
        query = "/" + str(uuid)
    else:
        query = "?parse="
        if site:
            # There is a known issue in tpdb, where site names are not matched due to casing.
            # example Teens3Some fails, but Teens3some succeeds.  Turns out Teens3Some is treated as 'Teens 3 Some'
            # and Teens3some is treated correctly as 'Teens 3some'.  Also, 'brazzersextra' still match 'Brazzers Extra'
            # Hense, the hack of lower casing the site.
            query += quote(re.sub(r"[^a-z0-9]", "", unidecode(site).lower())) + "."
        if release_date:
            query += release_date + "."
        if name:
            query += quote(re.sub(r" ", ".", name))
        query += "&limit=25"

    return f"{namer_config.override_tpdb_address}scenes{query}"


def __get_metadataapi_net_info(url: str, name_parts: FileNameParts, namer_config: NamerConfig):
    json_response = __get_response_json_object(url, namer_config)
    file_infos = []
    if json_response and json_response.strip() != "":
        logger.debug("json_response: \n{}", json_response)
        json_obj = json.loads(json_response, object_hook=lambda d: SimpleNamespace(**d))
        formatted = json.dumps(json.loads(json_response), indent=4, sort_keys=True)
        file_infos = __metadataapi_response_to_data(json_obj, url, formatted, name_parts)

    return file_infos


def __get_metadataapi_net_fileinfo(name_parts: FileNameParts, namer_config: NamerConfig, skip_date: bool, skip_name: bool) -> List[LookedUpFileInfo]:
    release_date = name_parts.date if not skip_date else None
    name = name_parts.name if not skip_name else None
    url = __build_url(namer_config, name_parts.site, release_date, name)
    file_infos = __get_metadataapi_net_info(url, name_parts, namer_config)

    return file_infos


def get_complete_metadatapi_net_fileinfo(name_parts: FileNameParts, uuid: str, namer_config: NamerConfig) -> Optional[LookedUpFileInfo]:
    url = __build_url(namer_config, uuid=uuid)
    file_infos = __get_metadataapi_net_info(url, name_parts, namer_config)
    if file_infos:
        return file_infos[0]

    return None


def match(file_name_parts: Optional[FileNameParts], namer_config: NamerConfig) -> ComparisonResults:
    """
    Give parsed file name parts, and a porndb token, returns a sorted list of possible matches.
    Matches will appear first.
    """
    if not file_name_parts:
        return ComparisonResults([])
    comparison_results = __metadata_api_lookup(file_name_parts, namer_config)
    comparison_results = sorted(comparison_results, key=__match_percent, reverse=True)
    # Works around the porndb not returning all info on search queries by looking up the full data
    # with the uuid of the best match.
    if comparison_results and comparison_results[0].is_match():
        uuid = comparison_results[0].looked_up.uuid
        if uuid:
            file_infos: LookedUpFileInfo | None = get_complete_metadatapi_net_fileinfo(file_name_parts, uuid, namer_config)
            if file_infos:
                comparison_results[0].looked_up = file_infos
    return ComparisonResults(comparison_results)


def main(args_list: List[str]):
    """
    Looks up metadata from metadataapi.net base on file name.
    """
    description = """
    Command line interface to look up a suggested name for an adult movie file based on an input string
    that is parsable by namer_file_parser.py
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("-c", "--configfile", help="override location for a configuration file.", type=Path)
    parser.add_argument("-f", "--file", help="File we want to provide a match name for.", required=True, type=Path)
    parser.add_argument("-j", "--jsonfile", help="write returned json to this file.", type=Path)
    parser.add_argument("-v", "--verbose", help="verbose, print logs", action="store_true")
    args = parser.parse_args(args=args_list)

    level = "DEBUG" if args.verbose else "ERROR"
    logger.remove()
    logger.add(sys.stdout, format="{time} {level} {message}", level=level)
    config = default_config()
    file_name = make_command(Path(args.file), config, ignore_file_restrictions=True)

    results: Optional[ComparisonResults] = None
    if file_name and file_name.parsed_file:
        results = match(file_name.parsed_file, config)

    if results:
        matched = results.get_match()
        if matched:
            print(matched.looked_up.new_file_name(config.inplace_name))
            if args.jsonfile and matched.looked_up and matched.looked_up.original_response:
                Path(args.jsonfile).write_text(matched.looked_up.original_response, encoding="UTF-8")


if __name__ == "__main__":
    main(sys.argv[1:])



from io import BytesIO
from typing import Optional

import requests
from loguru import logger
from requests_cache import CachedSession


class Http:
    @staticmethod
    def request(method, url, **kwargs):
        logger.info(f'Requesting {method} "{url}"')
        cache_session: Optional[CachedSession] = kwargs.get('cache_session')
        if 'cache_session' in kwargs:
            del kwargs['cache_session']

        if kwargs.get("stream", False) or not isinstance(cache_session, CachedSession):
            return requests.request(method, url, **kwargs)
        else:
            return cache_session.request(method, url, **kwargs)

    @staticmethod
    def get(url: str, **kwargs):
        return Http.request("GET", url, **kwargs)

    @staticmethod
    def post(url: str, **kwargs):
        return Http.request("POST", url, **kwargs)

    @staticmethod
    def head(url: str, **kwargs):
        return Http.request("HEAD", url, **kwargs)

    @staticmethod
    def download_file(url: str, **kwargs) -> Optional[BytesIO]:
        kwargs.setdefault('stream', True)
        http = Http.get(url, **kwargs)
        if http.ok:
            f = BytesIO()
            for data in http.iter_content(1024):
                f.write(data)

            return f



import re
import string


class PartialFormatter(string.Formatter):
    """
    Used for formatting NamerConfig.inplace_name and NamerConfig.
    """

    supported_keys = [
        "date",
        "description",
        "name",
        "site",
        "full_site",
        "performers",
        "all_performers",
        "act",
        "ext",
        "trans",
        "uuid",
    ]

    def __init__(self, missing="~~", bad_fmt="!!"):
        self.missing, self.bad_fmt = missing, bad_fmt

    def get_field(self, field_name, args, kwargs):
        # Handle a key not found
        try:
            val = super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError) as err:
            val = None, field_name
            if field_name not in self.supported_keys:
                raise KeyError(f"Key {field_name} not in support keys: {self.supported_keys}") from err
        return val

    def format_field(self, value, format_spec: str):
        if not value:
            return self.missing
        try:
            if re.match(r".\d+s", format_spec):
                value = value + format_spec[0] * int(format_spec[1:-1])
                format_spec = ""
            if re.match(r".\d+p", format_spec):
                value = format_spec[0] * int(format_spec[1:-1]) + value
                format_spec = ""
            if re.match(r".\d+i", format_spec):
                value = format_spec[0] * int(format_spec[1:-1]) + value + format_spec[0] * int(format_spec[1:-1])
                format_spec = ""
            return super().format_field(value, format_spec)
        except ValueError:
            if self.bad_fmt:
                return self.bad_fmt
            raise



"""
ffmpeg is access through this file, it is used to find the video stream's resolution,
and update audio streams "Default" setting.   Apple video players require there be
only one default audio stream, and this script lets you set it with the correct language
code if there are more than one audio streams and if they are correctly labeled.
See:  https://iso639-3.sil.org/code_tables/639/data/ for language codes.
"""
import json
from dataclasses import dataclass
import shutil
import string
from functools import lru_cache
from io import BytesIO
from pathlib import Path
from random import choices
from typing import Any, Dict, List, Optional

import ffmpeg
from loguru import logger
from PIL import Image


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class FFProbeStream:
    index: int                      # stream numer
    codec_name: str                 # "mp3", "h264", "hvec", "png"
    codec_type: str                 # "audio" or "video"
    disposition_default: bool       # default stream of this type
    disposition_attached_pic: bool  # is the "video" stream an attached picture.
    duration: float                 # seconds
    bit_rate: int                   # bitrate of the track
    # audio
    tags_language: Optional[str]            # 3 letters representing language of track (only matters for audio)
    # video only
    width: Optional[int] = None
    height: Optional[int] = None            # 720 1080 2160
    avg_frame_rate: Optional[float] = None  # average frames per second

    def __str__(self) -> str:
        data = self.to_dict()

        return json.dumps(data, indent=2)

    def to_dict(self) -> dict:
        data = {
            'codec_name': self.codec_name,
            'width': self.width,
            'height': self.height,
            'codec_type': self.codec_type,
            'framerate': self.avg_frame_rate,
            'duration': self.duration,
            'disposition_default': self.disposition_default,
        }

        return data

    def is_audio(self) -> bool:
        return self.codec_type == "audio"

    def is_video(self) -> bool:
        return self.codec_type == "video" and (not self.disposition_attached_pic or self.disposition_attached_pic is False)


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class FFProbeFormat:
    duration: float
    size: int
    bit_rate: int
    tags: Dict[str, str]


@dataclass(init=False, repr=False, eq=True, order=False, unsafe_hash=True, frozen=False)
class FFProbeResults:
    __results: List[FFProbeStream]
    __format: FFProbeFormat

    def __init__(self, data: List[FFProbeStream], probe_format: FFProbeFormat):
        self.__results = data
        self.__format = probe_format

    def get_default_video_stream(self) -> Optional[FFProbeStream]:
        for result in self.__results:
            if result.is_video() and result.disposition_default:
                return result

    def get_default_audio_stream(self) -> Optional[FFProbeStream]:
        for result in self.__results:
            if result.is_audio() and result.disposition_default:
                return result

    def get_audio_stream(self, language_code: str) -> Optional[FFProbeStream]:
        for result in self.__results:
            if result.is_audio() and result.tags_language == language_code:
                return result

    def get_all_streams(self) -> List[FFProbeStream]:
        return self.__results

    def get_format(self) -> FFProbeFormat:
        return self.__format


def get_resolution(file: Path) -> int:
    """
    Gets the vertical resolution of a mp4 file.  For example, 720, 1080, 2160...
    Returns zero if resolution can not be determined.
    """
    logger.info("resolution stream of file {}", file)
    probe = ffprobe(file)
    if probe:
        stream = probe.get_default_video_stream()
        if stream:
            return stream.height if stream.height else 0

    return 0


@logger.catch
def ffprobe(file: Path) -> Optional[FFProbeResults]:
    """
    Get the typed results of probing a video stream with ffprobe.
    """

    return _ffprobe(file, file.stat().st_size, file.stat().st_mtime)


@lru_cache(maxsize=1024)
def _ffprobe(file: Path, file_size: int, file_update: float) -> Optional[FFProbeResults]:
    """
    Get the typed results of probing a video stream with ffprobe.
    """

    logger.info("ffprobe file {}", file)
    ffprobe_out: Optional[Any] = None
    try:
        ffprobe_out = ffmpeg.probe(file)
    except:
        pass

    if not ffprobe_out:
        return

    streams = [stream for stream in ffprobe_out['streams'] if stream['codec_type'] in ('video', 'audio')]
    if not streams:
        return

    output: List[FFProbeStream] = []
    for stream in streams:
        ff_stream = FFProbeStream()
        ff_stream.bit_rate = int(stream['bit_rate']) if 'bit_rate' in stream else -1
        ff_stream.codec_name = stream['codec_name']
        ff_stream.codec_type = stream['codec_type']
        ff_stream.index = int(stream['index'])
        ff_stream.duration = float(stream['duration']) if 'duration' in stream else -1

        ff_stream.height = int(stream['height']) if 'height' in stream else -1
        ff_stream.width = int(stream['width']) if 'width' in stream else -1
        ff_stream.tags_language = stream['tags']['language'] if 'tags' in stream and 'language' in stream['tags'] else None

        if 'disposition' in stream:
            ff_stream.disposition_attached_pic = stream['disposition']['attached_pic'] == 1
            ff_stream.disposition_default = stream['disposition']['default'] == 1

        if 'avg_frame_rate' in stream:
            numer, denom = stream['avg_frame_rate'].split('/', 2)
            numer, denom = int(numer), int(denom)
            if numer != 0 and denom != 0:
                ff_stream.avg_frame_rate = numer / denom

        output.append(ff_stream)

    probe_format = FFProbeFormat()
    if 'format' in ffprobe_out:
        probe_format.bit_rate = int(ffprobe_out['format']['bit_rate'])
        probe_format.duration = float(ffprobe_out['format']['duration'])
        probe_format.size = int(ffprobe_out['format']['size'])
        probe_format.tags = ffprobe_out['format']['tags'] if 'tags' in ffprobe_out['format'] else {}

    return FFProbeResults(output, probe_format)


def get_audio_stream_for_lang(file: Path, language: str) -> int:
    """
    given a mp4 input file and a desired language will return the stream position of that language in the mp4.
    if the language is None, or the stream is not found, or the desired stream is the only default stream, None is returned.
    See: https://iso639-3.sil.org/code_tables/639/data/

    Returns -1 if stream can not be determined
    """

    stream_index = -1
    probe = ffprobe(file)
    if probe:
        stream = probe.get_audio_stream(language)
        if stream:
            stream_index = stream.index - 1 if not stream.disposition_default else -1

    return stream_index


def update_audio_stream_if_needed(mp4_file: Path, language: Optional[str]) -> bool:
    """
    Returns true if the file had to be edited to have a default audio stream equal to the desired language,
    mostly a concern for apple players (Quicktime/Apple TV/etc.)
    Copies, and potentially updates the default audio stream of a video file.
    """

    random = "".join(choices(population=string.ascii_uppercase + string.digits, k=10))
    temp_filename = f'{mp4_file.stem}_{random}' + mp4_file.suffix
    work_file = mp4_file.parent / temp_filename

    stream = get_audio_stream_for_lang(mp4_file, language) if language else None
    if stream and stream >= 0:
        process = (
            ffmpeg
            .input(mp4_file)
            .output(str(work_file), **{
                'map': 0,  # copy all stream
                'disposition:a': 'none',  # mark all audio streams as not default
                f'disposition:a:{stream}': 'default',  # mark this audio stream as default
                'c': 'copy'  # don't re-encode anything.
            })
            .run_async(quiet=True)
        )

        stdout, stderr = process.communicate()
        success = process.returncode == 0
        if not success:
            logger.info("Could not update audio stream for {}", mp4_file)
            if stderr:
                logger.info(stderr)
        else:
            logger.warning("Return code: {}", process.returncode)
            mp4_file.unlink()
            shutil.move(work_file, mp4_file)

        return success

    return True


def attempt_fix_corrupt(mp4_file: Path) -> bool:
    """
    Attempt to fix corrupt mp4 files.
    """
    random = "".join(choices(population=string.ascii_uppercase + string.digits, k=10))
    temp_filename = f'{mp4_file.stem}_{random}' + mp4_file.suffix
    work_file = mp4_file.parent / temp_filename

    logger.info("Attempt to fix damaged mp4 file: {}", mp4_file)
    process = (
        ffmpeg
        .input(mp4_file)
        .output(str(work_file), c='copy')
        .run_async(quiet=True)
    )

    stdout, stderr = process.communicate()
    success = process.returncode == 0
    if not success:
        logger.info("Could not fix mp4 files {}", mp4_file)
        if stderr:
            logger.info(stderr)
    else:
        logger.warning("Return code: {}", process.returncode)
        mp4_file.unlink()
        shutil.move(work_file, mp4_file)

    return success


def extract_screenshot(file: Path, time: float, screenshot_width: int = -1) -> Image.Image:
    out, _ = (
        ffmpeg
        .input(file, ss=time)
        .filter('scale', screenshot_width, -1)
        .output('pipe:', vframes=1, format='apng')
        .run(quiet=True, capture_stdout=True)
    )
    out = BytesIO(out)
    image = Image.open(out)

    return image
